"""Unit tests for ``rhodes.structures``."""
import pytest

from rhodes.structures import ContextPath

pytestmark = [pytest.mark.local, pytest.mark.functional]


_VALID_STATIC_CONTEXT_PATHS = (
    "$$",
    "$$.Execution",
    "$$.Execution.Id",
    "$$.Execution.StartTime",
    "$$.State",
    "$$.State.EnteredTime",
    "$$.State.Name",
    "$$.State.RetryCount",
    "$$.StateMachine",
    "$$.StateMachine.Id",
    "$$.Task",
    "$$.Task.Token",
    "$$.Map",
    "$$.Map.Item",
    "$$.Map.Item.Index",
    "$$.Map.Item.Value",
)
_VALID_CONTEXT_PATHS_WITH_INPUT = _VALID_STATIC_CONTEXT_PATHS + (
    "$$.Execution.Input",
    "$$.Execution.Input.foo",
    "$$.Execution.Input.foo.bar",
    "$$.Execution.Input.foo.bar.baz",
)


@pytest.mark.parametrize("path", _VALID_CONTEXT_PATHS_WITH_INPUT)
def test_contextpath_valid(path):
    ContextPath(path=path)


@pytest.mark.parametrize("path", _VALID_CONTEXT_PATHS_WITH_INPUT)
def test_contextpath_getattr_valid(path):
    expected = ContextPath(path=path)

    names = path.split(".")[1:]

    test = ContextPath()
    for name in names:
        test = getattr(test, name)

    assert test == expected


def test_contextpath_getattr_readable():
    """The real testing is via ``test_contextpath_getattr_valid``.

    This test is just to show a more human-readable form.
    """
    assert ContextPath() == ContextPath("$$")
    assert ContextPath().Execution == ContextPath("$$.Execution")
    assert ContextPath().Map.Item.Index == ContextPath("$$.Map.Item.Index")
    assert ContextPath().Execution.Input.foo.bar.baz == ContextPath("$$.Execution.Input.foo.bar.baz")


@pytest.mark.parametrize(
    "path",
    (pytest.param("", id="empty path"), pytest.param("$.Execution", id="valid child but invalid root"))
    + tuple(pytest.param(val + ".foo", id="valid prefix but invalid child") for val in _VALID_STATIC_CONTEXT_PATHS),
)
def test_contextpath_invalid(path):
    with pytest.raises(ValueError) as excinfo:
        ContextPath(path=path)

    excinfo.match("Invalid Context Path")



"""Amazon DynamoDB Integrations for Task states.

https://docs.aws.amazon.com/step-functions/latest/dg/connect-ddb.html
"""
import attr
from attr.validators import instance_of

from rhodes._util import RequiredValue
from rhodes.states.services import IntegrationPattern, ServiceArn, ServiceIntegration, _supports_patterns

__all__ = (
    "AmazonDynamoDb",
    "AmazonDynamoDbGetItem",
    "AmazonDynamoDbPutItem",
    "AmazonDynamoDbDeleteItem",
    "AmazonDynamoDbUpdateItem",
)

_ddb_supports_patterns = _supports_patterns(IntegrationPattern.REQUEST_RESPONSE)


class AmazonDynamoDb:
    TableName = attr.ib(validator=instance_of(str))

    def get_item(self, **kwargs):
        return AmazonDynamoDbGetItem(TableName=self.TableName, **kwargs)

    def put_item(self, **kwargs):
        return AmazonDynamoDbPutItem(TableName=self.TableName, **kwargs)

    def delete_item(self, **kwargs):
        return AmazonDynamoDbDeleteItem(TableName=self.TableName, **kwargs)

    def update_item(self, **kwargs):
        return AmazonDynamoDbUpdateItem(TableName=self.TableName, **kwargs)


def _ddb_table_name(cls: ServiceIntegration) -> ServiceIntegration:
    cls.TableName = attr.ib(default=None)

    return cls


def _ddb_key(cls: ServiceIntegration) -> ServiceIntegration:
    cls.Key = attr.ib(default=None)

    return cls


def _ddb_write_attributes(cls: ServiceIntegration) -> ServiceIntegration:
    cls.ConditionalOperator = attr.ib(default=None)
    cls.ConditionExpression = attr.ib(default=None)
    cls.Expected = attr.ib(default=None)
    cls.ExpressionAttributeNames = attr.ib(default=None)
    cls.ExpressionAttributeValues = attr.ib(default=None)
    cls.ReturnConsumedCapacity = attr.ib(default=None)
    cls.ReturnItemCollectionMetrics = attr.ib(default=None)
    cls.ReturnValues = attr.ib(default=None)

    return cls


@attr.s(eq=False)
@_ddb_key
@_ddb_table_name
@_ddb_supports_patterns
class AmazonDynamoDbGetItem(IntegrationPattern):
    _required_fields = (
        RequiredValue("Key", "Amazon DynamoDB GetItem Task requires a key value"),
        RequiredValue("TableName", "Amazon DynamoDB GetItem Task requires a table value"),
    )
    _resource_name = ServiceArn.DYNAMODB_GET_ITEM

    # TODO: Sort out validation rules
    #  https://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_GetItem.html#DDB-GetItem-request-Key

    AttributesToGet = attr.ib(default=None)
    ConsistentRead = attr.ib(default=None)
    ExpressionAttributeNames = attr.ib(default=None)
    ProjectionExpression = attr.ib(default=None)
    ReturnConsumedCapacity = attr.ib(default=None)


@attr.s(eq=False)
@_ddb_write_attributes
@_ddb_table_name
@_ddb_supports_patterns
class AmazonDynamoDbPutItem(IntegrationPattern):
    _required_fields = (
        RequiredValue("Item", "Amazon DynamoDB PutItem Task requires an item value"),
        RequiredValue("TableName", "Amazon DynamoDB PutItem Task requires a table value"),
    )
    _resource_name = ServiceArn.DYNAMODB_PUT_ITEM

    # TODO: Sort out validation rules
    #  https://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_PutItem.html#DDB-PutItem-request-Item

    Item = attr.ib(default=None)


@attr.s(eq=False)
@_ddb_write_attributes
@_ddb_key
@_ddb_table_name
@_ddb_supports_patterns
class AmazonDynamoDbDeleteItem(IntegrationPattern):
    _required_fields = (
        RequiredValue("Item", "Amazon DynamoDB DeleteItem Task requires a key value"),
        RequiredValue("TableName", "Amazon DynamoDB DeleteItem Task requires a table value"),
    )
    _resource_name = ServiceArn.DYNAMODB_DELETE_ITEM

    # TODO: Sort out validation rules
    #  https://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_DeleteItem.html#API_DeleteItem_RequestParameters


@attr.s(eq=False)
@_ddb_write_attributes
@_ddb_key
@_ddb_table_name
@_ddb_supports_patterns
class AmazonDynamoDbUpdateItem(IntegrationPattern):
    _required_fields = (
        RequiredValue("Item", "Amazon DynamoDB UpdateItem Task requires a key value"),
        RequiredValue("TableName", "Amazon DynamoDB UpdateItem Task requires a table value"),
    )
    _resource_name = ServiceArn.DYNAMODB_UPDATE_ITEM

    # TODO: Sort out validation rules
    #  https://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_UpdateItem.html#API_UpdateItem_RequestParameters

    AttributeUpdates = attr.ib(default=None)
    UpdateExpression = attr.ib(default=None)



"""Unit test suite for ``rhodes._converters``."""
from enum import Enum

import jsonpath_rw
import pytest

from rhodes._converters import convert_to_json_path
from rhodes.choice_rules import VariablePath
from rhodes.structures import JsonPath

pytestmark = [pytest.mark.local, pytest.mark.functional]


RAW_PATH = "$.path.to.value"


class MyEnum(Enum):
    MY_PATH = RAW_PATH


@pytest.mark.parametrize(
    "path_value",
    (
        pytest.param(RAW_PATH, id="string"),
        pytest.param(MyEnum.MY_PATH, id="enum"),
        pytest.param(jsonpath_rw.parse(RAW_PATH), id="jsonpath_rw.JSONPath"),
        pytest.param(JsonPath(RAW_PATH), id="JsonPath"),
        pytest.param(VariablePath(RAW_PATH), id="VariablePath"),
    ),
)
def test_convert_to_json_path(path_value):
    expected = JsonPath(RAW_PATH)

    test = convert_to_json_path(path_value)

    assert test.path == expected.path



"""AWS Batch Integrations for Task states.

https://docs.aws.amazon.com/step-functions/latest/dg/connect-batch.html
"""
from typing import Dict, Iterable

import attr
from attr.validators import deep_mapping, instance_of

from rhodes._util import RequiredValue
from rhodes.states.services import IntegrationPattern, ServiceArn, ServiceIntegration, _supports_patterns

__all__ = ("AwsBatch",)


@attr.s(eq=False)
@_supports_patterns(IntegrationPattern.REQUEST_RESPONSE, IntegrationPattern.SYNCHRONOUS)
class AwsBatch(ServiceIntegration):
    _required_fields = (
        RequiredValue("JobDefinition", "AWS Batch Task requires a job definition."),
        RequiredValue("JobName", "AWS Batch Task requires a job name."),
        RequiredValue("JobQueue", "AWS Batch Task requires a job queue."),
    )
    _resource_name = ServiceArn.BATCH

    # TODO: Sort out validation rules
    #  https://docs.aws.amazon.com/batch/latest/APIReference/API_SubmitJob.html#Batch-SubmitJob-request-arrayProperties

    # The ARN/name
    JobDefinition = attr.ib(default=None)
    # unique name...does SFn actually require this?
    JobName: str = attr.ib(default=None, validator=instance_of(str))
    # ??? ARN for something?
    JobQueue = attr.ib(default=None)
    # Arbitrary string-string map
    Parameters: Dict[str, str] = attr.ib(
        default=attr.Factory(dict),
        validator=deep_mapping(key_validator=instance_of(str), value_validator=instance_of(str)),
    )
    # TODO: These others feel like things that should already exist in Troposphere
    ArrayProperties: Iterable = attr.ib(default=None)
    ContainerOverrides = attr.ib(default=None)
    DependsOn = attr.ib(default=None)
    RetryStrategy = attr.ib(default=None)
    Timeout = attr.ib(default=None)



"""AWS Service Integrations for Task states.

https://docs.aws.amazon.com/step-functions/latest/dg/concepts-service-integrations.html
"""
from enum import Enum
from typing import Callable, Dict

import attr
from attr.validators import in_

from rhodes._util import require_field
from rhodes.states import Parameters, State, Task, task_type

__all__ = ("ServiceArn", "ServiceIntegration", "IntegrationPattern")

_DDB_BASE_ARN = "arn:aws:states:::dynamodb"
_SAGEMAKER_BASE_ARN = "arn:aws:states:::sagemaker"


class ServiceArn(Enum):
    AWSLAMBDA = "arn:aws:states:::lambda:invoke"
    BATCH = "arn:aws:states:::batch:submitJob"
    ECS = "arn:aws:states:::ecs:runTask"
    SNS = "arn:aws:states:::sns:publish"
    SQS = "arn:aws:states:::sqs:sendMessage"
    GLUE = "arn:aws:states:::glue:startJobRun"
    STEP_FUNCTIONS = "arn:aws:states:::states:startExecution"
    DYNAMODB_GET_ITEM = f"{_DDB_BASE_ARN}:getItem"
    DYNAMODB_PUT_ITEM = f"{_DDB_BASE_ARN}:putItem"
    DYNAMODB_DELETE_ITEM = f"{_DDB_BASE_ARN}:deleteItem"
    DYNAMODB_UPDATE_ITEM = f"{_DDB_BASE_ARN}:updateItem"
    SAGEMAKER_CREATE_ENDPOINT = f"{_SAGEMAKER_BASE_ARN}:createEndpoint"
    SAGEMAKER_CREATE_ENDPOINT_CONFIG = f"{_SAGEMAKER_BASE_ARN}:createEndpointConfig"
    SAGEMAKER_CREATE_HYPER_PARAMETER_TUNING_JOB = f"{_SAGEMAKER_BASE_ARN}:createHyperParameterTuningJob"
    SAGEMAKER_CREATE_LABELING_JOB = f"{_SAGEMAKER_BASE_ARN}:createLabelingJob"
    SAGEMAKER_CREATE_MODEL = f"{_SAGEMAKER_BASE_ARN}:createModel"
    SAGEMAKER_CREATE_TRAINING_JOB = f"{_SAGEMAKER_BASE_ARN}:createTrainingJob"
    SAGEMAKER_CREATE_TRANSFORM_JOB = f"{_SAGEMAKER_BASE_ARN}:createTransformJob"
    SAGEMAKER_UPDATE_ENDPOINT = f"{_SAGEMAKER_BASE_ARN}:updateEndpoint"


class IntegrationPattern(Enum):
    REQUEST_RESPONSE = ""
    SYNCHRONOUS = ".sync"
    WAIT_FOR_CALLBACK = ".waitForTaskToken"


@attr.s(eq=False)
@task_type
class ServiceIntegration(State):
    _resource_name = NotImplemented

    def to_task(self) -> Task:
        raise NotImplementedError()

    def to_dict(self) -> Dict:
        for required in self._required_fields:
            require_field(instance=self, required_value=required)

        task = self._build_task()
        return task.to_dict()

    def _build_task(self: "ServiceIntegration") -> Task:
        task_fields = [field.name for field in attr.fields(Task)]
        field_name_blacklist = ("Pattern", "Parameters")
        resource_name = self._resource_name.value + self.Pattern.value

        task_kwargs = {}
        parameters_kwargs = {}

        for field in attr.fields(type(self)):
            if field.name in field_name_blacklist or field.name.startswith("_"):
                continue

            value = getattr(self, field.name)
            if value is None:
                continue

            if field.name in task_fields:
                task_kwargs[field.name] = value
            else:
                parameters_kwargs[field.name] = value

        params = Parameters(**parameters_kwargs)
        return Task(Parameters=params, Resource=resource_name, **task_kwargs)


def _supports_patterns(*options: IntegrationPattern) -> Callable[[ServiceIntegration], ServiceIntegration]:
    def _decorate(cls: ServiceIntegration) -> ServiceIntegration:
        cls.Pattern: IntegrationPattern = attr.ib(default=options[0], validator=in_(options))

        return cls

    return _decorate



"""AWS Glue Integrations for Task states.

https://docs.aws.amazon.com/step-functions/latest/dg/connect-glue.html
"""
import attr

from rhodes._util import RequiredValue
from rhodes.states.services import IntegrationPattern, ServiceArn, ServiceIntegration, _supports_patterns


@attr.s(eq=False)
@_supports_patterns(IntegrationPattern.REQUEST_RESPONSE, IntegrationPattern.SYNCHRONOUS)
class AwsGlue(ServiceIntegration):
    _required_fields = ()
    _resource_name = ServiceArn.GLUE

    # TODO: Sort out validation rules
    #  https://docs.aws.amazon.com/step-functions/latest/dg/connect-glue.html

    JobName = attr.ib(default=None)
    JobRunId = attr.ib(default=None)
    Arguments = attr.ib(default=None)
    # TODO: AllocatedCapacity is deprecated; Will SFn's integration change?
    AllocatedCapacity = attr.ib(default=None)
    Timeout = attr.ib(default=None)
    SecurityConfiguration = attr.ib(default=None)
    NotificationProperty = attr.ib(default=None)



"""Simple state machine definition tests."""
import pytest

from rhodes import StateMachine, choice_rules
from rhodes.choice_rules import VariablePath
from rhodes.exceptions import InvalidDefinitionError
from rhodes.states import Choice, Fail, Map, Parallel, Pass, State, Succeed, Task, Wait

from ..unit_test_helpers import state_machine_body

pytestmark = [pytest.mark.local, pytest.mark.functional]


def test_hello_world():
    expected = state_machine_body("hello-world")

    test = StateMachine(
        Comment="A simple minimal example of the States language",
        StartAt="Hello World",
        States={
            "Hello World": Task(
                "Hello World", Resource="arn:aws:lambda:us-east-1:123456789012:function:HelloWorld", End=True
            )
        },
    )

    actual = test.to_dict()

    assert actual == expected


@pytest.mark.parametrize(
    "kwargs",
    (
        dict(),
        dict(States=[]),
        dict(States={"AlwaysBlue": Succeed("AlwaysBlue")}),
        dict(StartAt="AlwaysBlue"),
        dict(States={"AlwaysBlue": Succeed("AlwaysBlue")}, StartAt="Unknown"),
    ),
)
def test_required_values(kwargs):
    test = StateMachine(**kwargs)

    with pytest.raises(InvalidDefinitionError) as excinfo:
        test.to_dict()


def test_three_tasks():
    expected = state_machine_body("three-tasks")

    test = StateMachine(
        Comment="This is a state machine with three simple tasks.",
        StartAt="TaskOne",
        States=dict(
            TaskOne=Task("TaskOne", Resource="arn:aws:lambda:us-east-1:123456789012:function:One", Next="TaskTwo"),
            TaskTwo=Task("TaskTwo", Resource="arn:aws:lambda:us-east-1:123456789012:function:Two", Next="TaskThree"),
            TaskThree=Task("TaskThree", Resource="arn:aws:lambda:us-east-1:123456789012:function:Three", End=True),
        ),
    )

    actual = test.to_dict()

    assert actual == expected


def test_three_tasks_new_1():
    test = StateMachine(Comment="This is a state machine with three simple tasks.")

    task_one = test.start_with(Task("TaskOne", Resource="arn:aws:lambda:us-east-1:123456789012:function:One"))
    task_two = task_one.then(Task("TaskTwo", Resource="arn:aws:lambda:us-east-1:123456789012:function:Two"))
    task_three = task_two.then(Task("TaskThree", Resource="arn:aws:lambda:us-east-1:123456789012:function:Three"))
    task_three.end()

    expected = state_machine_body("three-tasks")
    actual = test.to_dict()
    assert actual == expected


def test_simple_choice():
    test = StateMachine(
        Comment="This is a simple state machine with a single choice and three end states.",
        StartAt="TheBeginning",
        States=dict(
            TheBeginning=Choice(
                "TheBeginning",
                Choices=[
                    choice_rules.StringEquals(Variable="$.value", Value="A", Next="ResultA"),
                    choice_rules.StringEquals(Variable="$.value", Value="B", Next="ResultB1"),
                ],
                Default="Unknown",
            ),
            ResultA=Task("ResultA", Resource="arn:aws:lambda:us-east-1:123456789012:function:A", End=True),
            ResultB1=Task("ResultB1", Resource="arn:aws:lambda:us-east-1:123456789012:function:B1", Next="ResultB2"),
            ResultB2=Task("ResultB2", Resource="arn:aws:lambda:us-east-1:123456789012:function:B2", End=True),
            Unknown=Fail("Unknown", Error="Unhandled Case", Cause="Unknown Value"),
        ),
    )

    expected = state_machine_body("simple-choice")
    actual = test.to_dict()
    assert actual == expected


def test_simple_choice_new_1():
    test = StateMachine(Comment="This is a simple state machine with a single choice and three end states.")
    result_a = test.add_state(Task("ResultA", Resource="arn:aws:lambda:us-east-1:123456789012:function:A", End=True))
    result_b1 = test.add_state(Task("ResultB1", Resource="arn:aws:lambda:us-east-1:123456789012:function:B1"))
    result_b1.then(Task("ResultB2", Resource="arn:aws:lambda:us-east-1:123456789012:function:B2", End=True))
    unknown = test.add_state(Fail("Unknown", Error="Unhandled Case", Cause="Unknown Value"))

    decision = test.start_with(Choice("TheBeginning"))
    decision.if_(choice_rules.StringEquals(Variable="$.value", Value="A")).then(result_a)
    decision.if_(choice_rules.StringEquals(Variable="$.value", Value="B")).then(result_b1)
    decision.else_(unknown)

    expected = state_machine_body("simple-choice")
    actual = test.to_dict()
    assert actual == expected


def test_simple_choice_new_2():
    test = StateMachine(Comment="This is a simple state machine with a single choice and three end states.")
    result_a = test.add_state(Task("ResultA", Resource="arn:aws:lambda:us-east-1:123456789012:function:A", End=True))
    result_b1 = test.add_state(Task("ResultB1", Resource="arn:aws:lambda:us-east-1:123456789012:function:B1"))
    result_b1.then(Task("ResultB2", Resource="arn:aws:lambda:us-east-1:123456789012:function:B2", End=True))
    unknown = test.add_state(Fail("Unknown", Error="Unhandled Case", Cause="Unknown Value"))

    decision = test.start_with(Choice("TheBeginning"))
    decision.if_(VariablePath("$.value") == "A").then(result_a)
    decision.if_(VariablePath("$.value") == "B").then(result_b1)
    decision.else_(unknown)

    expected = state_machine_body("simple-choice")
    actual = test.to_dict()
    assert actual == expected


def test_simple_choice_new_3():
    test = StateMachine(Comment="This is a simple state machine with a single choice and three end states.")
    result_a = Task("ResultA", Resource="arn:aws:lambda:us-east-1:123456789012:function:A", End=True)
    result_b1 = test.add_state(Task("ResultB1", Resource="arn:aws:lambda:us-east-1:123456789012:function:B1"))
    result_b1.then(Task("ResultB2", Resource="arn:aws:lambda:us-east-1:123456789012:function:B2", End=True))
    unknown = Fail("Unknown", Error="Unhandled Case", Cause="Unknown Value")

    decision = test.start_with(Choice("TheBeginning"))
    decision.if_(VariablePath("$.value") == "A").then(result_a)
    decision.if_(VariablePath("$.value") == "B").then(result_b1)
    decision.else_(unknown)

    expected = state_machine_body("simple-choice")
    actual = test.to_dict()
    assert actual == expected


def test_simple_choice_new_3_3():
    test = StateMachine(Comment="This is a simple state machine with a single choice and three end states.")
    result_a = Task("ResultA", Resource="arn:aws:lambda:us-east-1:123456789012:function:A").end()
    result_b1 = test.add_state(Task("ResultB1", Resource="arn:aws:lambda:us-east-1:123456789012:function:B1"))
    result_b1.then(Task("ResultB2", Resource="arn:aws:lambda:us-east-1:123456789012:function:B2").end())
    unknown = Fail("Unknown", Error="Unhandled Case", Cause="Unknown Value")

    decision = test.start_with(Choice("TheBeginning"))
    decision.if_(VariablePath("$.value") == "A").then(result_a)
    decision.if_(VariablePath("$.value") == "B").then(result_b1)
    decision.else_(unknown)

    expected = state_machine_body("simple-choice")
    actual = test.to_dict()
    assert actual == expected


def test_simple_choice_new_4():
    test = StateMachine(Comment="This is a simple state machine with a single choice and three end states.")

    decision = test.start_with(Choice("TheBeginning"))
    decision.if_(VariablePath("$.value") == "A").then(
        Task("ResultA", Resource="arn:aws:lambda:us-east-1:123456789012:function:A").end()
    )
    result_b1 = decision.if_(VariablePath("$.value") == "B").then(
        Task("ResultB1", Resource="arn:aws:lambda:us-east-1:123456789012:function:B1")
    )
    result_b1.then(Task("ResultB2", Resource="arn:aws:lambda:us-east-1:123456789012:function:B2").end())
    decision.else_(Fail("Unknown", Error="Unhandled Case", Cause="Unknown Value"))

    expected = state_machine_body("simple-choice")
    actual = test.to_dict()
    assert actual == expected


def test_simple_parallel():
    expected = state_machine_body("simple-parallel")

    test = StateMachine(
        Comment="Parallel Example.",
        StartAt="LookupCustomerInfo",
        States=dict(
            LookupCustomerInfo=Parallel(
                "LookupCustomerInfo",
                End=True,
                Branches=[
                    StateMachine(
                        StartAt="LookupAddress",
                        States=dict(
                            LookupAddress=Task(
                                "LookupAddress",
                                Resource="arn:aws:lambda:us-east-1:123456789012:function:AddressFinder",
                                End=True,
                            )
                        ),
                    ),
                    StateMachine(
                        StartAt="LookupPhone",
                        States=dict(
                            LookupPhone=Task(
                                "LookupPhone",
                                Resource="arn:aws:lambda:us-east-1:123456789012:function:PhoneFinder",
                                End=True,
                            )
                        ),
                    ),
                ],
            )
        ),
    )

    actual = test.to_dict()

    assert actual == expected


def test_simple_parallel_new_1():
    lookup_address = StateMachine()
    lookup_address.start_with(
        Task("LookupAddress", Resource="arn:aws:lambda:us-east-1:123456789012:function:AddressFinder")
    ).end()

    lookup_phone = StateMachine()
    lookup_phone.start_with(
        Task("LookupPhone", Resource="arn:aws:lambda:us-east-1:123456789012:function:PhoneFinder")
    ).end()

    test = StateMachine(Comment="Parallel Example.")
    test.start_with(Parallel("LookupCustomerInfo", Branches=[lookup_address, lookup_phone])).end()

    expected = state_machine_body("simple-parallel")
    actual = test.to_dict()
    assert actual == expected


def test_simple_parallel_new_2():
    lookup_address = StateMachine()
    lookup_address.start_with(
        Task("LookupAddress", Resource="arn:aws:lambda:us-east-1:123456789012:function:AddressFinder")
    ).end()

    lookup_phone = StateMachine()
    lookup_phone.start_with(
        Task("LookupPhone", Resource="arn:aws:lambda:us-east-1:123456789012:function:PhoneFinder")
    ).end()

    parallel_run = Parallel("LookupCustomerInfo")
    parallel_run.add_branch(lookup_address)
    parallel_run.add_branch(lookup_phone)

    test = StateMachine(Comment="Parallel Example.")
    test.start_with(parallel_run).end()

    expected = state_machine_body("simple-parallel")
    actual = test.to_dict()
    assert actual == expected


def test_simple_map():
    test = StateMachine(
        Comment="Simple state machine with one map state",
        StartAt="ValidateAll",
        States=dict(
            ValidateAll=Map(
                "ValidateAll",
                InputPath="$.detail",
                ItemsPath="$.shipped",
                MaxConcurrency=0,
                Iterator=StateMachine(
                    StartAt="Validate",
                    States=dict(
                        Validate=Task(
                            "Validate", Resource="arn:aws:lambda:us-east-1:123456789012:function:ship-val", End=True
                        )
                    ),
                ),
                ResultPath="$.detail.shipped",
                End=True,
            )
        ),
    )

    expected = state_machine_body("simple-map")
    actual = test.to_dict()
    assert actual == expected


def test_simple_map_new_1():
    validate_task = Task("Validate", Resource="arn:aws:lambda:us-east-1:123456789012:function:ship-val")

    state_iterator = StateMachine()
    state_iterator.start_with(validate_task).end()

    mapper = Map(
        "ValidateAll",
        InputPath="$.detail",
        ItemsPath="$.shipped",
        MaxConcurrency=0,
        Iterator=state_iterator,
        ResultPath="$.detail.shipped",
    )

    test = StateMachine(Comment="Simple state machine with one map state")
    test.start_with(mapper).end()

    expected = state_machine_body("simple-map")
    actual = test.to_dict()
    assert actual == expected



from datetime import datetime
from decimal import Decimal
from enum import Enum

import attr
from attr.validators import deep_iterable, instance_of, optional

from ._converters import convert_to_json_path
from .exceptions import InvalidDefinitionError
from .structures import JsonPath


class VariablePath(JsonPath):
    """JsonPath with overloading helper methods."""

    # TODO: Add __and__ and __or__ behaviors?

    def __lt__(self, other):
        return derive_rule(variable=self, operator="<", value=other)

    def __le__(self, other):
        return derive_rule(variable=self, operator="<=", value=other)

    def __eq__(self, other):
        return derive_rule(variable=self, operator="==", value=other)

    def __ne__(self, other):
        inner_rule = derive_rule(variable=self, operator="==", value=other)
        return Not(Rule=inner_rule)

    def __gt__(self, other):
        return derive_rule(variable=self, operator=">", value=other)

    def __ge__(self, other):
        return derive_rule(variable=self, operator=">=", value=other)


def _required_next(instance):
    if instance.Next is None:
        raise InvalidDefinitionError("ChoiceRule missing state transition")


def _require_choice_rule_instance(*, class_name: str, attribute_name: str, value):
    if not isinstance(value, ChoiceRule):
        raise TypeError(f'"{class_name}.{attribute_name}" must be a "ChoiceRule". Received "{type(value)}"')


def _require_no_next(*, class_name: str, attribute_name: str, value):
    if value.Next is not None:
        raise ValueError(f'"{class_name}.{attribute_name}" must not have a "Next" value defined.')


def _single_to_dict(instance, suppress_next=False):
    if not suppress_next:
        _required_next(instance)

    instance_dict = {instance.__class__.__name__: instance._serialized_value(), "Variable": str(instance.Variable)}
    if instance.Next is not None:
        instance_dict["Next"] = instance.Next

    return instance_dict


def _convert_to_variable_path(value) -> VariablePath:
    if isinstance(value, VariablePath):
        return value

    return VariablePath(value)


def _single(cls):
    cls.Variable = attr.ib(validator=instance_of(VariablePath), converter=_convert_to_variable_path)
    cls.Next = attr.ib(default=None, validator=optional(instance_of(str)))
    cls.to_dict = _single_to_dict

    return cls


def _multi_to_dict(instance, suppress_next=False):
    if not suppress_next:
        _required_next(instance)

    # TODO: Validate that no children have a Next value

    return {
        instance.__class__.__name__: [rule.to_dict(suppress_next=True) for rule in instance.Rules],
        "Next": instance.Next,
    }


def _validate_multi_subrules(instance, attribute, value):
    for pos, rule in enumerate(value):
        position_name = f"{attribute.name}[{pos}]"
        _require_choice_rule_instance(class_name=instance.__class__.__name__, attribute_name=position_name, value=rule)
        _require_no_next(class_name=instance.__class__.__name__, attribute_name=position_name, value=rule)


def _multi(cls):
    cls.Rules = attr.ib(validator=_validate_multi_subrules)
    cls.Next = attr.ib(default=None, validator=optional(instance_of(str)))
    cls.to_dict = _multi_to_dict

    return cls


def _string(cls):
    cls.Value = attr.ib(validator=instance_of(str))
    cls = _single(cls)

    return cls


def _number(cls):
    def _numeric_converter(value) -> Decimal:
        if isinstance(value, Decimal):
            return value

        return Decimal(str(value))

    def _value_serializer(instance) -> float:
        return float(instance.Value)

    # TODO: Note that for interoperability,
    #  numeric comparisons should not be assumed to work
    #  with values outside the magnitude or precision
    #  representable using the IEEE 754-2008 “binary64” data type.
    #  In particular,
    #  integers outside of the range [-(253)+1, (253)-1]
    #  might fail to compare in the expected way.
    cls.Value = attr.ib(validator=instance_of(Decimal), converter=_numeric_converter)
    cls._serialized_value = _value_serializer
    cls = _single(cls)

    return cls


def _bool(cls):
    cls.Value = attr.ib(validator=instance_of(bool))
    cls = _single(cls)

    return cls


def _timestamp(cls):
    def _datetime_validator(instance, attribute, value):
        if value.tzinfo is None:
            raise ValueError(f"'{attribute.name}' must have a 'tzinfo' value set.")

    def _value_serializer(instance):
        return instance.Value.isoformat()

    cls.Value = attr.ib(validator=[instance_of(datetime), _datetime_validator])
    cls._serialized_value = _value_serializer
    cls = _single(cls)

    return cls


@attr.s(eq=False)
class ChoiceRule:

    member_of = None

    def __eq__(self, other: "ChoiceRule") -> bool:
        if not isinstance(other, self.__class__):
            return False

        if not self.to_dict() == other.to_dict():
            return False

        if self.member_of != other.member_of:
            return False

        return True

    def __ne__(self, other: "ChoiceRule") -> bool:
        return not self.__eq__(other)

    def _serialized_value(self):
        return self.Value

    def then(self, state):
        if self.Next is not None:
            raise InvalidDefinitionError(f"Choice rule already has a defined target")

        self.member_of.member_of.add_state(state)

        self.Next = state.name

        return state


@attr.s(eq=False)
@_string
class StringEquals(ChoiceRule):
    pass


@attr.s(eq=False)
@_string
class StringLessThan(ChoiceRule):
    pass


@attr.s(eq=False)
@_string
class StringGreaterThan(ChoiceRule):
    pass


@attr.s(eq=False)
@_string
class StringLessThanEquals(ChoiceRule):
    pass


@attr.s(eq=False)
@_string
class StringGreaterThanEquals(ChoiceRule):
    pass


@attr.s(eq=False)
@_number
class NumericEquals(ChoiceRule):
    pass


@attr.s(eq=False)
@_number
class NumericLessThan(ChoiceRule):
    pass


@attr.s(eq=False)
@_number
class NumericGreaterThan(ChoiceRule):
    pass


@attr.s(eq=False)
@_number
class NumericLessThanEquals(ChoiceRule):
    pass


@attr.s(eq=False)
@_number
class NumericGreaterThanEquals(ChoiceRule):
    pass


@attr.s(eq=False)
@_bool
class BooleanEquals(ChoiceRule):
    pass


@attr.s(eq=False)
@_timestamp
class TimestampEquals(ChoiceRule):
    pass


@attr.s(eq=False)
@_timestamp
class TimestampLessThan(ChoiceRule):
    pass


@attr.s(eq=False)
@_timestamp
class TimestampGreaterThan(ChoiceRule):
    pass


@attr.s(eq=False)
@_timestamp
class TimestampLessThanEquals(ChoiceRule):
    pass


@attr.s(eq=False)
@_timestamp
class TimestampGreaterThanEquals(ChoiceRule):
    pass


@attr.s(eq=False)
@_multi
class And(ChoiceRule):
    pass


@attr.s(eq=False)
@_multi
class Or(ChoiceRule):
    pass


@attr.s(eq=False)
class Not(ChoiceRule):
    Rule = attr.ib(validator=instance_of(ChoiceRule))
    Next = attr.ib(default=None, validator=optional(instance_of(str)))

    @Rule.validator
    def _validate_rule(self, attribute, value):
        _require_choice_rule_instance(class_name=self.__class__.__name__, attribute_name=attribute.name, value=value)
        _require_no_next(class_name=self.__class__.__name__, attribute_name=attribute.name, value=value)

    def to_dict(self, suppress_next=False):
        if not suppress_next:
            _required_next(self)

        inner_rule = self.Rule.to_dict(suppress_next=True)
        instance_dict = dict(Not=inner_rule)

        if self.Next is not None:
            instance_dict["Next"] = self.Next

        return instance_dict


_OPERATORS = {
    "string": {
        "==": StringEquals,
        "<": StringLessThan,
        "<=": StringLessThanEquals,
        ">": StringGreaterThan,
        ">=": StringGreaterThanEquals,
    },
    "number": {
        "==": NumericEquals,
        "<": NumericLessThan,
        "<=": NumericLessThanEquals,
        ">": NumericGreaterThan,
        ">=": NumericGreaterThanEquals,
    },
    "time": {
        "==": TimestampEquals,
        "<": TimestampLessThan,
        "<=": TimestampLessThanEquals,
        ">": TimestampGreaterThan,
        ">=": TimestampGreaterThanEquals,
    },
    "boolean": {"==": BooleanEquals},
}
_TYPE_MAP = {bool: "boolean", int: "number", float: "number", Decimal: "number", str: "string", datetime: "time"}


def derive_rule(*, variable: VariablePath, operator: str, value) -> ChoiceRule:
    if isinstance(value, Enum):
        value = value.value

    try:
        value_type = _TYPE_MAP[type(value)]
    except KeyError:
        raise TypeError(f'Unhandled value type "{type(value)}"')

    try:
        operator_class = _OPERATORS[value_type][operator]
    except KeyError:
        raise ValueError(f'Unhandled operator "{operator}"')

    return operator_class(Variable=variable, Value=value)


def all_(*rules: ChoiceRule) -> And:
    return And(Rules=list(rules))


def any_(*rules: ChoiceRule) -> Or:
    return Or(Rules=list(rules))



"""Converter functions for use with ``attrs`` attributes."""
from enum import Enum
from typing import Optional, Union

import jsonpath_rw

from .structures import JsonPath

__all__ = ("convert_to_json_path",)


def convert_to_json_path(value: Optional[Union[str, Enum, jsonpath_rw.JSONPath, JsonPath]]) -> Union[None, JsonPath]:
    """Converter for any attributes that must be :class:`JsonPath` instances."""
    if value is None:
        return value

    if isinstance(value, JsonPath):
        return value

    if isinstance(value, Enum):
        value = value.value

    return JsonPath(value)



"""Amazon SNS Integrations for Task states.

https://docs.aws.amazon.com/step-functions/latest/dg/connect-sns.html
"""
import attr

from rhodes._util import RequiredValue
from rhodes.states.services import IntegrationPattern, ServiceArn, ServiceIntegration, _supports_patterns


@attr.s(eq=False)
@_supports_patterns(IntegrationPattern.REQUEST_RESPONSE, IntegrationPattern.WAIT_FOR_CALLBACK)
class AmazonSns(ServiceIntegration):
    _required_fields = (RequiredValue("Message", "Amazon SNS Task requires a message"),)
    _resource_name = ServiceArn.SNS

    # TODO: Sort out validation rules
    #  https://docs.aws.amazon.com/sns/latest/api/API_Publish.html#API_Publish_RequestParameters

    Message = attr.ib(default=None)
    MessageAttributes = attr.ib(default=None)
    MessageStructure = attr.ib(default=None)
    Subject = attr.ib(default=None)
    # one of the below is required
    PhoneNumber = attr.ib(default=None)
    TargetArn = attr.ib(default=None)
    TopicArn = attr.ib(default=None)



"""Other structures for Rhodes."""
from typing import Union

import attr
import jsonpath_rw
from attr.validators import instance_of

__all__ = ("JsonPath", "ContextPath")


def _convert_path(value: Union[str, jsonpath_rw.JSONPath, "JsonPath"]) -> jsonpath_rw.JSONPath:
    if isinstance(value, jsonpath_rw.JSONPath):
        return value

    if isinstance(value, JsonPath):
        return value.path

    return jsonpath_rw.parse(value)


@attr.s
class JsonPath:
    """Represents a JSONPath variable in request/response body."""

    path: jsonpath_rw.JSONPath = attr.ib(validator=instance_of(jsonpath_rw.JSONPath), converter=_convert_path)

    def __str__(self):
        return str(self.path)


@attr.s
class ContextPath:
    """Represents a JSONPath(ish) variable in the Context Object.

    https://docs.aws.amazon.com/step-functions/latest/dg/input-output-contextobject.html
    """

    _path: str = attr.ib(default="$$")

    @_path.validator
    def _validate_path(self, attribute, value):
        _valid_paths = {
            "Execution": {"Id": False, "Input": True, "StartTime": False},
            "State": {"EnteredTime": False, "Name": False, "RetryCount": False},
            "StateMachine": {"Id": False},
            "Task": {"Token": False},
            "Map": {"Item": {"Index": False, "Value": False}},
        }

        def _validate_parts(tree, fields):
            # retrieve relative root
            key, *remaining = fields

            # look for valid children
            try:
                inner_tree = tree[key]
            except KeyError:
                raise ValueError("Invalid Context Path")

            if isinstance(inner_tree, dict):
                if not remaining:
                    # Requested path is a sub-object not an individual member
                    return

                return _validate_parts(inner_tree, remaining)

            if not inner_tree and remaining:
                # Requested path is an unknown child
                raise ValueError("Invalid Context Path")

        # TODO: Verify that:
        #  1. Aside from the leading $, path is a valid JSONPath
        #  2. If the prefix is anything other than $$.Execution.Input,
        #   the path must be on the known list.

        parts = value.split(".")

        # Path MUST start with "$$." not "$." to identify as Context Object
        if parts[0] != "$$":
            raise ValueError("Invalid Context Path")

        if len(parts) == 1:
            # Requested path is the entire Context Object
            return

        _validate_parts(_valid_paths, parts[1:])

    def __str__(self):
        return self._path

    def __getattr__(self, item):
        return ContextPath(f"{self._path}.{item}")



"""Complex state machine definition tests."""
import pytest

from rhodes import StateMachine, choice_rules
from rhodes.choice_rules import VariablePath, all_
from rhodes.states import Choice, Fail, Parallel, Parameters, Succeed, Task, Wait
from rhodes.structures import JsonPath

from ..unit_test_helpers import compare_state_machine

pytestmark = [pytest.mark.local, pytest.mark.functional]

PARSE_REQUIREMENTS_RESOURCE = "arn:aws:lambda:us-east-1:123456789012:function:parse-requirements"
BUILD_PYTHON_36_RESOURCE = "arn:aws:lambda:us-east-1:123456789012:function:build-py36"
BUILD_PYTHON_37_RESOURCE = "arn:aws:lambda:us-east-1:123456789012:function:build-py37"
EVENT_FILTER_RESOURCE = "arn:aws:lambda:us-east-1:123456789012:function:event-filter"
ARTIFACT_LOCATOR_RESOURCE = "arn:aws:lambda:us-east-1:123456789012:function:artifact-locator"
LAYER_VERSION_PUBLISHER_RESOURCE = "arn:aws:lambda:us-east-1:123456789012:function:layer-version-publisher"
NOTIFY_TOPIC = "arn:aws:sns:us-east-1:123456789012:accretion-notify"


def test_accretion_builder():
    test = StateMachine(
        Comment="Artifact Builder",
        StartAt="ParseRequirements",
        States=dict(
            ParseRequirements=Task("ParseRequirements", Resource=PARSE_REQUIREMENTS_RESOURCE, Next="SelectLanguage"),
            SelectLanguage=Choice(
                "SelectLanguage",
                Choices=[choice_rules.StringEquals(Variable="$.Language", Value="python", Next="BuildPython")],
                Default="UnknownLanguage",
            ),
            UnknownLanguage=Fail("UnknownLanguage", Cause="Invalid language"),
            BuildPython=Parallel(
                "BuildPython",
                Branches=[
                    StateMachine(
                        StartAt="BuildPython36",
                        States=dict(BuildPython36=Task("BuildPython36", Resource=BUILD_PYTHON_36_RESOURCE, End=True)),
                    ),
                    StateMachine(
                        StartAt="BuildPython37",
                        States=dict(BuildPython37=Task("BuildPython37", Resource=BUILD_PYTHON_37_RESOURCE, End=True)),
                    ),
                ],
                ResultPath="$.BuildResults",
                End=True,
            ),
        ),
    )

    compare_state_machine("accretion_builder", test)


def test_accretion_builder_new_1():
    parse_requirements = Task("ParseRequirements", Resource=PARSE_REQUIREMENTS_RESOURCE)

    build_python = Parallel("BuildPython", ResultPath="$.BuildResults")

    build_python_36 = build_python.add_branch()
    build_python_36.start_with(Task("BuildPython36", Resource=BUILD_PYTHON_36_RESOURCE)).end()

    build_python_37 = build_python.add_branch()
    build_python_37.start_with(Task("BuildPython37", Resource=BUILD_PYTHON_37_RESOURCE)).end()

    unknown_language = Fail("UnknownLanguage", Cause="Invalid language")

    test = StateMachine(Comment="Artifact Builder")
    select_language = test.start_with(parse_requirements).then(Choice("SelectLanguage"))

    # TODO: Auto-add children to parent if they were added before the choice was added to parent
    # TODO: Add Choice.elseif_() ?
    select_language.if_(VariablePath("$.Language") == "python").then(build_python)
    select_language.else_(unknown_language)

    build_python.end()

    compare_state_machine("accretion_builder", test)


def test_accretion_listener():

    test = StateMachine(
        Comment="Replication Listener",
        StartAt="Filter",
        States={
            "Filter": Task("Filter", Resource=EVENT_FILTER_RESOURCE, ResultPath="$", Next="ShouldProcess"),
            "ShouldProcess": Choice(
                "ShouldProcess",
                Choices=[choice_rules.BooleanEquals(Variable="$.ProcessEvent", Value=True, Next="LocateArtifact")],
                Default="IgnoreEvent",
            ),
            "IgnoreEvent": Succeed("IgnoreEvent", Comment="Ignore this event"),
            "LocateArtifact": Task(
                "LocateArtifact", Resource=ARTIFACT_LOCATOR_RESOURCE, ResultPath="$.Artifact", Next="ArtifactCheck"
            ),
            "ArtifactCheck": Choice(
                "ArtifactCheck",
                Choices=[
                    choice_rules.BooleanEquals(Variable="$.Artifact.Found", Value=True, Next="PublishNewVersion"),
                    choice_rules.And(
                        Rules=[
                            choice_rules.BooleanEquals(Variable="$.Artifact.Found", Value=False),
                            choice_rules.NumericGreaterThan(Variable="$.Artifact.ReadAttempts", Value=15),
                        ],
                        Next="ReplicationTimeout",
                    ),
                ],
                Default="WaitForReplication",
            ),
            "ReplicationTimeout": Fail("ReplicationTimeout", Error="Timed out waiting for artifact to replicate"),
            "WaitForReplication": Wait("WaitForReplication", Seconds=60, Next="LocateArtifact"),
            "PublishNewVersion": Task(
                "PublishNewVersion", Resource=LAYER_VERSION_PUBLISHER_RESOURCE, ResultPath="$.Layer", Next="Notify"
            ),
            "Notify": Task(
                "Notify",
                Resource="arn:aws:states:::sns:publish",
                Parameters=Parameters(**{"TopicArn": NOTIFY_TOPIC, "Message.$": "$.Layer"}),
                End=True,
            ),
        },
    )

    compare_state_machine("accretion_listener", test)


def test_accretion_listener_new_1():

    test = StateMachine(Comment="Replication Listener")

    event_filter = test.start_with(Task("Filter", Resource=EVENT_FILTER_RESOURCE, ResultPath="$"))
    skip_check = event_filter.then(Choice("ShouldProcess"))
    skip_check.else_(Succeed("IgnoreEvent", Comment="Ignore this event"))

    locate_artifact = skip_check.if_(VariablePath("$.ProcessEvent") == True).then(
        Task("LocateArtifact", Resource=ARTIFACT_LOCATOR_RESOURCE, ResultPath="$.Artifact")
    )
    artifact_check = locate_artifact.then(Choice("ArtifactCheck"))

    publisher = artifact_check.if_(VariablePath("$.Artifact.Found") == True).then(
        Task("PublishNewVersion", Resource=LAYER_VERSION_PUBLISHER_RESOURCE, ResultPath="$.Layer")
    )
    publisher.then(
        Task(
            "Notify",
            Resource="arn:aws:states:::sns:publish",
            Parameters=Parameters(TopicArn=NOTIFY_TOPIC, Message=JsonPath("$.Layer")),
        )
    ).end()

    artifact_check.if_(
        all_(VariablePath("$.Artifact.Found") == False, VariablePath("$.Artifact.ReadAttempts") > 15)
    ).then(Fail("ReplicationTimeout", Error="Timed out waiting for artifact to replicate"))

    waiter = artifact_check.else_(Wait("WaitForReplication", Seconds=60))
    waiter.then(locate_artifact)

    compare_state_machine("accretion_listener", test)



""""""
import json
from typing import Any, Dict, Optional

import attr
from attr.validators import deep_iterable, deep_mapping, instance_of, optional
from troposphere import GetAtt, Ref, Sub, awslambda, stepfunctions

from rhodes._converters import convert_to_json_path
from rhodes._util import RequiredValue, require_field
from rhodes._validators import is_valid_arn, is_valid_timestamp
from rhodes.choice_rules import ChoiceRule
from rhodes.exceptions import InvalidDefinitionError
from rhodes.structures import ContextPath, JsonPath

__all__ = (
    "State",
    "StateMachine",
    "Pass",
    "Parameters",
    "Parallel",
    "Map",
    "Choice",
    "Task",
    "Wait",
    "Fail",
    "Succeed",
)


def _getatt_arn(value: str) -> str:
    return f"${{{value}.Arn}}"


def _ref(value: str) -> str:
    return f"${{{value}}}"


def _serialize_troposphere_value(value):
    # Inject appropriate Ref/GetAtt for Troposphere
    if isinstance(value, awslambda.Function):
        return _getatt_arn(value.title)

    if isinstance(value, stepfunctions.Activity):
        return _ref(value.title)

    if isinstance(value, Ref):
        return _ref(value.data["Ref"])

    if isinstance(value, GetAtt):
        return _getatt_arn(value.data["Fn::GetAtt"][0])

    return value


def _serialize_name_and_value(*, name: str, value: Any) -> [str, Any]:
    value = _serialize_troposphere_value(value)

    if isinstance(value, (JsonPath, ContextPath)):
        return name, str(value)

    if hasattr(value, "to_dict") and callable(value.to_dict):
        return name, value.to_dict()

    return name, value


@attr.s(eq=False)
class State:
    """"""

    name = attr.ib(validator=instance_of(str))
    member_of = None
    _required_fields = []
    __setup_complete = False

    def __attrs_post_init__(self):
        self.Type = self.__class__.__name__
        self.__setup_complete = True

    def __setattr__(self, name, value):
        """Validate the value attribute value after setting."""
        super(State, self).__setattr__(name, value)
        if self.__setup_complete:
            # Don't validate until after attrs is done setting up.
            attr.validate(self)

    def __eq__(self, other: "State") -> bool:
        if not isinstance(other, self.__class__):
            # TODO: What about the other direction?
            #  ex: Child() == Parent() vs Parent() == Child()
            return False

        if self.to_dict() != other.to_dict():
            return False

        if self.member_of != other.member_of:
            return False

        return True

    def __ne__(self, other: "State") -> bool:
        return not self.__eq__(other)

    def to_dict(self) -> Dict:
        for required in self._required_fields:
            require_field(instance=self, required_value=required)

        self_dict = {"Type": self.Type}
        for field in attr.fields(type(self)):
            if field.name == "name":
                continue

            value = getattr(self, field.name)
            if value is None:
                continue

            new_name, new_value = _serialize_name_and_value(name=field.name, value=value)

            self_dict[new_name] = new_value

        return self_dict


@attr.s
class StateMachine:
    """"""

    _required_fields = [
        RequiredValue("States", "State machine contains no states."),
        RequiredValue("StartAt", "State machine has no starting point."),
    ]
    __setup_complete = False

    # TODO: Map of States
    States = attr.ib(
        default=attr.Factory(dict),
        validator=deep_mapping(key_validator=instance_of(str), value_validator=instance_of(State)),
    )
    # TODO: Name of State
    StartAt = attr.ib(default=None, validator=optional(instance_of(str)))
    Comment = attr.ib(default=None, validator=optional(instance_of(str)))
    # TODO: MUST be 1.0 if provided
    Version = attr.ib(default=None, validator=optional(instance_of(str)))
    # TODO: MUST be non-negative
    TimeoutSeconds = attr.ib(default=None, validator=optional(instance_of(int)))

    def __attrs_post_init__(self):
        self.__setup_complete = True

    def __setattr__(self, name, value):
        """Validate the value attribute value after setting."""
        super(StateMachine, self).__setattr__(name, value)
        if self.__setup_complete:
            # Don't validate until after attrs is done setting up.
            attr.validate(self)

    def to_dict(self) -> Dict:
        for required in self._required_fields:
            require_field(instance=self, required_value=required)

        if self.StartAt not in self.States:
            raise InvalidDefinitionError(f"Starting state {self.StartAt!r} not in states {self.States.keys()!r}.")

        self_dict = dict(StartAt=self.StartAt)

        for optional_attribute in ("Comment", "TimeoutSeconds", "Version"):
            if getattr(self, optional_attribute) is not None:
                self_dict[optional_attribute] = getattr(self, optional_attribute)

        self_dict["States"] = {key: value.to_dict() for key, value in self.States.items()}

        return self_dict

    def definition_string(self) -> Sub:
        data = self.to_dict()
        initial_value = json.dumps(data)
        return Sub(initial_value)

    def add_state(self, new_state: State) -> State:
        if new_state.name in self.States:
            if self.States[new_state.name] == new_state:
                return new_state

            raise InvalidDefinitionError(f"State {new_state.name!r} already in state machine.")

        new_state.member_of = self
        # TODO: use references rather than extracting names
        self.States[new_state.name] = new_state
        return new_state

    def start_with(self, first_state: State) -> State:
        self.add_state(new_state=first_state)

        # TODO: use references rather than extracting names
        self.StartAt = first_state.name

        return first_state


def _comment(cls):
    cls.Comment = attr.ib(default=None, validator=optional(instance_of(str)))

    return cls


def _input_output(cls):
    cls.InputPath = attr.ib(default=None, validator=optional(instance_of(JsonPath)), converter=convert_to_json_path)
    cls.OutputPath = attr.ib(default=None, validator=optional(instance_of(JsonPath)), converter=convert_to_json_path)

    return cls


class Parameters:
    def __init__(self, **kwargs):
        self._map = kwargs

    def to_dict(self) -> Dict[str, Any]:
        def _inner():
            for name, value in self._map.items():
                new_name, new_value = _serialize_name_and_value(name=name, value=value)

                if isinstance(value, JsonPath) or isinstance(value, ContextPath):
                    # If you manually provide path strings, you must manually set the parameter suffix.
                    if not new_name.endswith(".$"):
                        new_name += ".$"

                yield new_name, new_value

        return dict(_inner())

    def __repr__(self):
        return f"{self.__class__.__name__}({', '.join(f'{name}={value!r}' for name, value in self._map.items())})"


def _parameters(cls):
    cls.Parameters = attr.ib(default=None, validator=optional(instance_of(Parameters)))

    return cls


def _result_path(cls):
    cls.ResultPath = attr.ib(default=None, validator=optional(instance_of(JsonPath)), converter=convert_to_json_path)

    return cls


def _next_or_end(cls):
    # TODO: These need to be required and exclusive OR
    cls.Next = attr.ib(default=None, validator=optional(instance_of(str)))
    # TODO: End=False is invalid
    cls.End = attr.ib(default=None, validator=optional(instance_of(bool)))

    def _set_next(instance: State, next_state: State) -> State:
        if instance.End:
            raise InvalidDefinitionError(
                "Cannot set state transition." f"State {instance.name!r} already has an end condition."
            )

        instance.member_of.add_state(next_state)
        # TODO: set reference rather than extracting name
        instance.Next = next_state.name
        return next_state

    def _set_end(instance: State) -> State:
        if instance.Next:
            raise InvalidDefinitionError(
                "Cannot set end condition." f"State {instance.name!r} already has a state transition."
            )

        instance.End = True

        return instance

    cls.then = _set_next
    cls.end = _set_end

    return cls


def _catch_retry(cls):
    cls.Retry = attr.ib(default=None)
    cls.Catch = attr.ib(default=None)

    return cls


@attr.s(eq=False)
@_comment
@_input_output
@_parameters
@_result_path
@_next_or_end
class Pass(State):
    Result = attr.ib(default=None)


def task_type(cls):
    cls = _catch_retry(cls)
    cls = _next_or_end(cls)
    cls = _result_path(cls)
    cls = _input_output(cls)
    cls = _comment(cls)

    # TODO: Timeout MUST be positive
    cls.TimeoutSeconds = attr.ib(default=None, validator=optional(instance_of(int)))
    # TODO: HeartbeatSeconds MUST be positive
    cls.HeartbeatSeconds = attr.ib(default=None, validator=optional(instance_of(int)))

    return cls


@attr.s(eq=False)
@_parameters
@task_type
class Task(State):
    _required_fields = [RequiredValue("Resource", "Task resource is not set.")]

    # TODO: Additional validation for strings?
    Resource = attr.ib(
        default=None, validator=optional(instance_of((str, awslambda.Function, stepfunctions.Activity, GetAtt, Ref)))
    )


@attr.s(eq=False)
@_comment
@_input_output
class Choice(State):
    # TODO: Validate that Next and Default states are in parent
    Choices = attr.ib(default=attr.Factory(list), validator=deep_iterable(member_validator=instance_of(ChoiceRule)))
    Default = attr.ib(default=None)

    def add_choice(self, rule: ChoiceRule) -> ChoiceRule:
        if rule.member_of is not None:
            if rule.member_of is self:
                return rule

            raise InvalidDefinitionError("Rule already added to another Choice state")

        self.Choices.append(rule)
        rule.member_of = self
        return rule

    def if_(self, rule: ChoiceRule) -> ChoiceRule:
        return self.add_choice(rule)

    def else_(self, state: State) -> State:
        if self.Default is not None:
            raise InvalidDefinitionError(f'Choice state "{self.name}" already has a Default transition.')

        self.member_of.add_state(state)

        # TODO: use references rather than extracting names
        self.Default = state.name

        return state

    def to_dict(self) -> Dict:
        if not self.Choices:
            raise InvalidDefinitionError(f'Choice state "{self.name}" has no defined choices.')

        self_dict = super(Choice, self).to_dict()

        for pos, branch in enumerate(self_dict["Choices"]):
            self_dict["Choices"][pos] = branch.to_dict()

        return self_dict


@attr.s(eq=False)
@_comment
@_input_output
@_next_or_end
class Wait(State):
    # TODO: Exactly one of these
    Seconds = attr.ib(default=None, validator=optional(instance_of(int)))
    # TODO: Timestamp must be ISO8601 timestamp
    Timestamp = attr.ib(default=None)
    SecondsPath = attr.ib(default=None, validator=optional(instance_of(JsonPath)), converter=convert_to_json_path)
    TimestampPath = attr.ib(default=None, validator=optional(instance_of(JsonPath)), converter=convert_to_json_path)

    @Timestamp.validator
    def _check_timestamp(self, attribute, value):
        if not is_valid_timestamp(value):
            raise ValueError("Invalid Timestamp value: {}".format(value))


@attr.s(eq=False)
@_comment
@_input_output
class Succeed(State):
    """"""


@attr.s(eq=False)
@_comment
class Fail(State):
    """"""

    Error = attr.ib(default=None, validator=optional(instance_of(str)))
    Cause = attr.ib(default=None, validator=optional(instance_of(str)))


@attr.s(eq=False)
@_comment
@_input_output
@_parameters
@_result_path
@_next_or_end
@_catch_retry
class Parallel(State):
    # TODO: Each branch MUST be a self-contained state machine.
    Branches = attr.ib(default=attr.Factory(list))

    def to_dict(self) -> Dict:
        self_dict = super(Parallel, self).to_dict()

        for pos, branch in enumerate(self_dict["Branches"]):
            self_dict["Branches"][pos] = branch.to_dict()

        return self_dict

    def add_branch(self, state_machine: Optional[StateMachine] = None):
        if state_machine is None:
            state_machine = StateMachine()

        self.Branches.append(state_machine)
        return state_machine


@attr.s(eq=False)
@_comment
@_input_output
@_parameters
@_result_path
@_next_or_end
@_catch_retry
class Map(State):
    # TODO: Iterator MUST be a self-contained state machine.
    Iterator = attr.ib(validator=instance_of(StateMachine))
    # TODO: ItemsPath MUST be a valid JSON-path
    ItemsPath = attr.ib(validator=optional(instance_of(JsonPath)), converter=convert_to_json_path)
    # TODO: MaxConcurrency MUST be non-negative
    MaxConcurrency = attr.ib(default=None, validator=optional(instance_of(int)))



"""Helpers for unit test suites."""
import json
from enum import Enum
from functools import partial
from pathlib import Path
from typing import Dict

import jsonpath_rw
import pytest

from rhodes.structures import JsonPath

__all__ = (
    "load_and_test_vectors",
    "state_machine_body",
    "state_body",
    "choice_rule_body",
    "compare_state_machine",
    "compare_state",
    "compare_choice_rule",
    "path_converter",
)
HERE = Path(__file__).parent
VECTORS_DIR = HERE / ".." / "vectors"


@pytest.fixture(params=(lambda x: x, JsonPath, jsonpath_rw.parse))
def path_converter(request):
    return request.param


class VectorTypes(Enum):
    def __init__(self, friendly_name: str, directory: Path):
        self.friendly_name = friendly_name
        self.directory = directory

    STATE_MACHINE = ("state machine", VECTORS_DIR / "state-machines")
    STATE = ("state", VECTORS_DIR / "states")
    CHOICE_RULE = ("choice rule", VECTORS_DIR / "choice-rules")


def _get_vector(vector_type: VectorTypes, vector_name: str) -> Dict:
    filepath = vector_type.directory / f"{vector_name}.json"

    if not filepath.is_file():
        raise FileNotFoundError(f"Unable to locate {vector_type.friendly_name} vector {vector_name}.json")

    with filepath.open("r") as vector:
        return json.load(vector)


state_machine_body = partial(_get_vector, VectorTypes.STATE_MACHINE)
state_body = partial(_get_vector, VectorTypes.STATE)
choice_rule_body = partial(_get_vector, VectorTypes.CHOICE_RULE)


def load_and_test_vectors(loader, *, kind, name, value):
    all_expected = loader(kind)
    expected = all_expected[name]

    actual = value.to_dict()

    assert actual == expected


def _assert_equal_to_vector(vector_type: VectorTypes, vector_name: str, value: Dict):
    expected = _get_vector(vector_type, vector_name)

    actual = value.to_dict()
    assert actual == expected

    actual_json = json.dumps(actual, indent=4, sort_keys=True)
    expected_json = json.dumps(expected, indent=4, sort_keys=True)

    assert actual_json == expected_json


compare_state_machine = partial(_assert_equal_to_vector, VectorTypes.STATE_MACHINE)
compare_state = partial(_assert_equal_to_vector, VectorTypes.STATE)
compare_choice_rule = partial(_assert_equal_to_vector, VectorTypes.CHOICE_RULE)



"""Amazon ECS/Fargate Integrations for Task states.

https://docs.aws.amazon.com/step-functions/latest/dg/connect-ecs.html
"""
import attr

from rhodes._util import RequiredValue
from rhodes.states.services import IntegrationPattern, ServiceArn, ServiceIntegration, _supports_patterns


@attr.s(eq=False)
@_supports_patterns(
    IntegrationPattern.REQUEST_RESPONSE, IntegrationPattern.SYNCHRONOUS, IntegrationPattern.WAIT_FOR_CALLBACK
)
class AmazonEcs(ServiceIntegration):
    _required_fields = (RequiredValue("TaskDefinition", "Amazon ECS Task requires a task definition"),)
    _resource_name = ServiceArn.ECS

    # TODO: Sort out validation rules
    #  https://docs.aws.amazon.com/AmazonECS/latest/APIReference/API_RunTask.html#ECS-RunTask-request-cluster

    Cluster = attr.ib(default=None)
    Group = attr.ib(default=None)
    LaunchType = attr.ib(default=None)
    NetworkConfiguration = attr.ib(default=None)
    Overrides = attr.ib(default=None)
    PlacementConstraints = attr.ib(default=None)
    PlacementStrategy = attr.ib(default=None)
    PlatformVersion = attr.ib(default=None)
    TaskDefinition = attr.ib(default=None)



"""AWS Lambda Integrations for Task states.

https://docs.aws.amazon.com/step-functions/latest/dg/connect-lambda.html
"""
import attr
from attr.validators import in_

from rhodes._util import RequiredValue
from rhodes.states.services import IntegrationPattern, ServiceArn, ServiceIntegration, _supports_patterns

__all__ = ("AwsLambda",)

AWS_LAMBDA_DEFAULT_INVOCATION_TYPE = "RequestResponse"
AWS_LAMBDA_INVOCATION_TYPES = ("Event", AWS_LAMBDA_DEFAULT_INVOCATION_TYPE, "DryRun")


@attr.s(eq=False)
@_supports_patterns(IntegrationPattern.REQUEST_RESPONSE, IntegrationPattern.WAIT_FOR_CALLBACK)
class AwsLambda(ServiceIntegration):
    _required_fields = (RequiredValue("FunctionName", "AWS Lambda Task requires a function name."),)
    _resource_name = ServiceArn.AWSLAMBDA

    # TODO: FunctionName MUST have length 1 <= n <= 170
    #  Pattern: (arn:(aws[a-zA-Z-]*)?:lambda:)?([a-z]{2}(-gov)?-[a-z]+-\d{1}:)?(\d{12}:)?(function:)?([a-zA-Z0-9-_\.]+)(:(\$LATEST|[a-zA-Z0-9-_]+))?
    FunctionName = attr.ib(default=None)
    # TODO: Determine validation for Payload
    #  Payload can also de-reference...how many layers deep?
    Payload = attr.ib(default=None)
    # TODO: ClientContext MUST have length n <= 3583
    ClientContext = attr.ib(default=None)
    InvocationType = attr.ib(default=AWS_LAMBDA_DEFAULT_INVOCATION_TYPE, validator=in_(AWS_LAMBDA_INVOCATION_TYPES))
    # TODO: Qualifier MUST have length 1 <= n <= 128
    Qualifier = attr.ib(default=None)



"""Utilities for use in Rhodes."""
from typing import Any, Callable

import attr

from .exceptions import InvalidDefinitionError


@attr.s(auto_attribs=True)
class RequiredValue:
    field_name: str
    error_message: str


def require_field(*, instance: Any, required_value: RequiredValue, validator: Callable[[Any], bool] = bool):
    # TODO: The validator here is incorrect.
    #  I need to pull the validatidator for the attribute from the instance class.
    if not hasattr(instance, required_value.field_name):
        raise InvalidDefinitionError(f"Field {required_value.field_name!r} missing.")

    if not validator(getattr(instance, required_value.field_name)):
        raise InvalidDefinitionError(required_value.error_message)



"""Unit tests for ``rhodes.choice_rules``."""
import operator
from datetime import datetime
from decimal import Decimal
from functools import partial

import pytest
import pytz

from rhodes.choice_rules import (
    And,
    BooleanEquals,
    ChoiceRule,
    Not,
    NumericEquals,
    NumericGreaterThan,
    NumericGreaterThanEquals,
    NumericLessThan,
    NumericLessThanEquals,
    Or,
    StringEquals,
    StringGreaterThan,
    StringGreaterThanEquals,
    StringLessThan,
    StringLessThanEquals,
    TimestampEquals,
    TimestampGreaterThan,
    TimestampGreaterThanEquals,
    TimestampLessThan,
    TimestampLessThanEquals,
    VariablePath,
    all_,
    any_,
)

from ..unit_test_helpers import choice_rule_body, load_and_test_vectors

pytestmark = [pytest.mark.local, pytest.mark.functional]
_load_and_test_vector = partial(load_and_test_vectors, choice_rule_body)

_OPERATOR_MAP = dict(
    GreaterThan=operator.gt,
    GreaterThanEquals=operator.ge,
    LessThan=operator.lt,
    LessThanEquals=operator.le,
    Equals=operator.eq,
)


@pytest.mark.parametrize("name, override_kwargs", (("CheckTrue", dict(Value=True)), ("CheckFalse", dict(Value=False))))
def test_boolean_good(name, override_kwargs):
    kwargs = dict(Next="NextState", Variable="$.value")
    kwargs.update(override_kwargs)

    test = BooleanEquals(**kwargs)

    _load_and_test_vector(kind="boolean", name=name, value=test)


def _timestamp_values():
    for cls in (
        TimestampEquals,
        TimestampGreaterThan,
        TimestampGreaterThanEquals,
        TimestampLessThan,
        TimestampLessThanEquals,
    ):
        base_name = cls.__name__.replace("Timestamp", "")
        yield (f"{base_name}Date", cls, dict(Value=pytz.utc.localize(datetime(2016, 12, 1))))
        yield (
            f"{base_name}DateTime",
            cls,
            dict(Value=pytz.timezone("US/Pacific").localize(datetime(2016, 12, 1, 15, 7, 7))),
        )


@pytest.mark.parametrize("name, cls, override_kwargs", _timestamp_values())
def test_timestamp_good(name, cls, override_kwargs):
    kwargs = dict(Next="NextState", Variable="$.value")
    kwargs.update(override_kwargs)

    test = cls(**kwargs)

    _load_and_test_vector(kind="timestamp", name=name, value=test)


def _timestamp_variable_statements():
    for name, cls, override_kwargs in _timestamp_values():
        base_name = cls.__name__.replace("Timestamp", "")
        yield (name, _OPERATOR_MAP[base_name], override_kwargs["Value"])


@pytest.mark.parametrize("name, op, value", _timestamp_variable_statements())
def test_timestamp_good_with_variable(name, op, value):
    test = op(VariablePath("$.value"), value)
    # TODO: Change this to use ChoiceRule.then() once they can exist outside of a parent
    test.Next = "NextState"

    _load_and_test_vector(kind="timestamp", name=name, value=test)


@pytest.mark.parametrize(
    "override_kwargs",
    (
        pytest.param(dict(Value="2016-12-01"), id="date string"),
        pytest.param(dict(Value=datetime(2016, 12, 1)), id="timezone-naive datetime"),
    ),
)
@pytest.mark.parametrize(
    "cls",
    (TimestampEquals, TimestampGreaterThan, TimestampGreaterThanEquals, TimestampLessThan, TimestampLessThanEquals),
)
def test_timestamp_invalid(cls, override_kwargs):
    kwargs = dict(Next="NextState", Variable="$.value")
    kwargs.update(override_kwargs)

    with pytest.raises(Exception):
        cls(**kwargs)


@pytest.mark.parametrize(
    "name, cls, override_kwargs",
    (
        ("Equals", StringEquals, dict(Value="Lorem ipsum dolor sit amet")),
        ("GreaterThan", StringGreaterThan, dict(Value="Lorem ipsum dolor sit amet")),
        ("GreaterThanEquals", StringGreaterThanEquals, dict(Value="Lorem ipsum dolor sit amet")),
        ("LessThan", StringLessThan, dict(Value="Lorem ipsum dolor sit amet")),
        ("LessThanEquals", StringLessThanEquals, dict(Value="Lorem ipsum dolor sit amet")),
    ),
)
def test_string_good(name, cls, override_kwargs):
    kwargs = dict(Next="NextState", Variable="$.value")
    kwargs.update(override_kwargs)

    test = cls(**kwargs)

    _load_and_test_vector(kind="string", name=name, value=test)


@pytest.mark.parametrize("override_kwargs", (dict(Value=42),))
@pytest.mark.parametrize(
    "cls", (StringEquals, StringGreaterThan, StringGreaterThanEquals, StringLessThan, StringLessThanEquals)
)
def test_string_invalid(cls, override_kwargs):
    kwargs = dict(Next="NextState", Variable="$.value")
    kwargs.update(override_kwargs)

    with pytest.raises(Exception):
        cls(**kwargs)


def _numeric_values():
    for cls in (NumericEquals, NumericLessThan, NumericLessThanEquals, NumericGreaterThan, NumericGreaterThanEquals):
        base_name = cls.__name__.replace("Numeric", "")
        for value in (42, 42.0, "42", "42.0", Decimal("42"), Decimal("42.0")):
            yield (base_name, cls, dict(Value=value))


@pytest.mark.parametrize("name, cls, override_kwargs", _numeric_values())
def test_numeric_good(name, cls, override_kwargs):
    kwargs = dict(Next="NextState", Variable="$.value")
    kwargs.update(override_kwargs)

    test = cls(**kwargs)

    _load_and_test_vector(kind="numeric", name=name, value=test)


@pytest.mark.parametrize("override_kwargs", (dict(Value="Lorem ipsum dolor sit amet"),))
@pytest.mark.parametrize(
    "cls", (NumericEquals, NumericGreaterThan, NumericGreaterThanEquals, NumericLessThan, NumericLessThanEquals)
)
def test_numeric_invalid(cls, override_kwargs):
    kwargs = dict(Next="NextState", Variable="$.value")
    kwargs.update(override_kwargs)

    with pytest.raises(Exception):
        cls(**kwargs)


def test_not_rule_good():
    test = Not(Rule=StringEquals(Variable="$.value", Value="Lorem ipsum dolor sit amet"), Next="NextState")

    _load_and_test_vector(kind="not", name="Simple", value=test)


def test_not_rule_good_with_variable():
    var = VariablePath("$.value")
    test = var != "Lorem ipsum dolor sit amet"
    # TODO: Change this to use ChoiceRule.then() once they can exist outside of a parent
    test.Next = "NextState"

    _load_and_test_vector(kind="not", name="Simple", value=test)


def test_and_rule_good():
    test = And(
        Rules=[NumericLessThan(Variable="$.value", Value=30), NumericGreaterThan(Variable="$.value", Value=20)],
        Next="NextState",
    )

    _load_and_test_vector(kind="and", name="Simple", value=test)


def test_and_rule_good_with_variable():
    var = VariablePath("$.value")
    test = And(Rules=[var < 30, var > 20])
    # TODO: Change this to use ChoiceRule.then() once they can exist outside of a parent
    test.Next = "NextState"

    _load_and_test_vector(kind="and", name="Simple", value=test)


def test_and_rule_good_with_variable_1():
    var = VariablePath("$.value")
    test = all_(var < 30, var > 20)
    # TODO: Change this to use ChoiceRule.then() once they can exist outside of a parent
    test.Next = "NextState"

    _load_and_test_vector(kind="and", name="Simple", value=test)


def test_and_many_rule_good_with_variable_1():
    var = VariablePath("$.value")
    test = all_(var < 30, var > 20, var != 22, var != 27)
    # TODO: Change this to use ChoiceRule.then() once they can exist outside of a parent
    test.Next = "NextState"

    _load_and_test_vector(kind="and", name="Many", value=test)


def test_or_rule_good():
    test = Or(
        Rules=[NumericLessThan(Variable="$.value", Value=20), NumericGreaterThan(Variable="$.value", Value=30)],
        Next="NextState",
    )

    _load_and_test_vector(kind="or", name="Simple", value=test)


def test_or_rule_good_with_variable():
    var = VariablePath("$.value")

    test = Or(Rules=[var < 20, var > 30])
    # TODO: Change this to use ChoiceRule.then() once they can exist outside of a parent
    test.Next = "NextState"

    _load_and_test_vector(kind="or", name="Simple", value=test)


def test_or_rule_good_with_variable_1():
    var = VariablePath("$.value")

    test = any_(var < 20, var > 30)
    # TODO: Change this to use ChoiceRule.then() once they can exist outside of a parent
    test.Next = "NextState"

    _load_and_test_vector(kind="or", name="Simple", value=test)


def test_or_many_rule_good_with_variable_1():
    var = VariablePath("$.value")

    test = any_(var < 20, var > 30, var == 22, var == 27)
    # TODO: Change this to use ChoiceRule.then() once they can exist outside of a parent
    test.Next = "NextState"

    _load_and_test_vector(kind="or", name="Many", value=test)



# pylint: disable=invalid-name
"""Sphinx configuration."""
import io
import os
import re
from datetime import datetime

VERSION_RE = re.compile(r"""__version__ = ['"]([0-9b.]+)['"]""")
HERE = os.path.abspath(os.path.dirname(__file__))


def read(*args):
    """Read complete file contents."""
    return io.open(os.path.join(HERE, *args), encoding="utf-8").read()


def get_release():
    """Read the release (full three-part version number) from this module."""
    init = read("..", "src", "rhodes", "__init__.py")
    return VERSION_RE.search(init).group(1)


def get_version():
    """Read the version (MAJOR.MINOR) from this module."""
    _release = get_release()
    split_version = _release.split(".")
    if len(split_version) == 3:
        return ".".join(split_version[:2])
    return _release


project = u"rhodes"
version = get_version()
release = get_release()

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_autodoc_typehints",
    "sphinxcontrib.spelling",
]
napoleon_include_special_with_doc = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

source_suffix = ".rst"  # The suffix of source filenames.
master_doc = "index"  # The master toctree document.

copyright = u"%s, Amazon" % datetime.now().year  # pylint: disable=redefined-builtin

# List of directories, relative to source directory, that shouldn't be searched
# for source files.
exclude_trees = ["_build"]

pygments_style = "sphinx"

autoclass_content = "both"
autodoc_default_flags = ["show-inheritance", "members"]
autodoc_member_order = "bysource"

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
htmlhelp_basename = "%sdoc" % project

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {"python": ("http://docs.python.org/", None)}

# autosummary
autosummary_generate = True

# Spellchecker
spelling_word_list_filename = "spelling_wordlist.txt"
spelling_lang = "en_US"
spelling_ignore_python_builtins = True
spelling_ignore_importable_modules = True



"""Unit test suite for ``rhodes.states.services.sns``."""
import pytest

from rhodes.states import Parameters, Task
from rhodes.states.services import IntegrationPattern
from rhodes.states.services.sns import AmazonSns
from rhodes.structures import JsonPath

pytestmark = [pytest.mark.local, pytest.mark.functional]


def test_amazon_sns_request_response():
    expected_task = Task(
        "TestTask",
        Resource="arn:aws:states:::sns:publish",
        Parameters=Parameters(
            TopicArn="arn:aws:sns:us-east-1:123456789012:accretion-notify", Message=JsonPath("$.foo.bar")
        ),
    )
    expected = expected_task.to_dict()

    test_task = AmazonSns(
        "TestTask", TopicArn="arn:aws:sns:us-east-1:123456789012:accretion-notify", Message=JsonPath("$.foo.bar")
    )
    test = test_task.to_dict()

    assert test == expected


def test_amazon_sns_callback():
    expected_task = Task(
        "TestTask",
        Resource="arn:aws:states:::sns:publish.waitForTaskToken",
        Parameters=Parameters(
            TopicArn="arn:aws:sns:us-east-1:123456789012:accretion-notify", Message=JsonPath("$.foo.bar")
        ),
    )
    expected = expected_task.to_dict()

    test_task = AmazonSns(
        "TestTask",
        Pattern=IntegrationPattern.WAIT_FOR_CALLBACK,
        TopicArn="arn:aws:sns:us-east-1:123456789012:accretion-notify",
        Message=JsonPath("$.foo.bar"),
    )
    test = test_task.to_dict()

    assert test == expected



"""Unit tests for ``rhodes.states``."""
from datetime import datetime
from functools import partial
from typing import Dict

import pytest
import pytz

from rhodes import StateMachine, choice_rules
from rhodes.states import Choice, Fail, Map, Parallel, Pass, State, Succeed, Task, Wait
from rhodes.structures import JsonPath

from ..unit_test_helpers import load_and_test_vectors, path_converter, state_body

pytestmark = [pytest.mark.local, pytest.mark.functional]
_load_and_test_vector = partial(load_and_test_vectors, state_body)


def test_pass(path_converter):
    name = "No-op"
    test = Pass(
        name,
        Result={"x-datum": 0.381018, "y-datum": 622.2269926397355},
        ResultPath=path_converter("$.coords"),
        Next="End",
    )

    _load_and_test_vector(kind="pass", name=name, value=test)


def test_task():
    name = "TaskState"
    test = Task(
        name,
        Comment="Task State example",
        Resource="arn:aws:states:us-east-1:123456789012:task:HelloWorld",
        TimeoutSeconds=300,
        HeartbeatSeconds=60,
        Next="NextState",
    )

    _load_and_test_vector(kind="task", name=name, value=test)


def test_choice(path_converter):
    name = "ChoiceStateX"
    test = Choice(
        name,
        Choices=[
            choice_rules.Not(
                Rule=choice_rules.StringEquals(Variable=path_converter("$.type"), Value="Private"), Next="Public"
            ),
            choice_rules.And(
                Rules=[
                    choice_rules.NumericGreaterThanEquals(Variable=path_converter("$.value"), Value=20),
                    choice_rules.NumericLessThan(Variable=path_converter("$.value"), Value=30),
                ],
                Next="ValueInTwenties",
            ),
            choice_rules.BooleanEquals(Variable=path_converter("$.value"), Value=True, Next="TrueState"),
            choice_rules.TimestampLessThan(
                Variable=path_converter("$.value"),
                Value=pytz.utc.localize(datetime(1999, 9, 13, 13, 0, 21)),
                Next="PartyTime",
            ),
        ],
        Default="DefaultState",
    )

    _load_and_test_vector(kind="choice", name=name, value=test)


@pytest.mark.parametrize(
    "name,kwargs",
    (
        ("wait_ten_seconds", dict(Seconds=10, Next="NextState")),
        ("wait_for_seconds_path", dict(SecondsPath="$.seconds", Next="NextState")),
        ("wait_for_seconds_path", dict(SecondsPath=JsonPath("$.seconds"), Next="NextState")),
        ("wait_until_timestamp", dict(Timestamp="2016-03-14T01:59:00Z", Next="NextState")),
        ("wait_until_timestamp_path", dict(TimestampPath="$.expirydate", Next="NextState")),
        ("wait_until_timestamp_path", dict(TimestampPath=JsonPath("$.expirydate"), Next="NextState")),
    ),
)
def test_wait(name: str, kwargs: Dict):
    test = Wait(name, **kwargs)

    _load_and_test_vector(kind="wait", name=name, value=test)


def test_succeed():
    name = "SuccessState"
    test = Succeed(name)

    _load_and_test_vector(kind="succeed", name=name, value=test)


def test_fail():
    name = "FailState"
    test = Fail(name, Error="ErrorA", Cause="Kaiju attack")

    _load_and_test_vector(kind="fail", name=name, value=test)


def test_parallel():
    name = "FunWithMath"
    test = Parallel(
        name,
        Branches=[
            StateMachine(StartAt="Add", States={"Add": Task("Add", Resource="arn:aws:states:::task:Add", End=True)}),
            StateMachine(
                StartAt="Subtract",
                States={"Subtract": Task("Subtract", Resource="arn:aws:states:::task:Subtract", End=True)},
            ),
        ],
        Next="NextState",
    )

    _load_and_test_vector(kind="parallel", name=name, value=test)


def test_parallel_new_1():
    name = "FunWithMath"

    branch_1 = StateMachine()
    branch_1.start_with(Task("Add", Resource="arn:aws:states:::task:Add")).end()
    branch_2 = StateMachine()
    branch_2.start_with(Task("Subtract", Resource="arn:aws:states:::task:Subtract")).end()

    test = Parallel(name)
    test.add_branch(branch_1)
    test.add_branch(branch_2)
    test.Next = "NextState"

    _load_and_test_vector(kind="parallel", name=name, value=test)


def test_map(path_converter):
    name = "Validate-All"
    test = Map(
        name,
        InputPath=path_converter("$.detail"),
        ItemsPath=path_converter("$.shipped"),
        MaxConcurrency=0,
        Iterator=StateMachine(
            StartAt="Validate",
            States={
                "Validate": Task(
                    "Validate", Resource="arn:aws:lambda:us-east-1:123456789012:function:ship-val", End=True
                )
            },
        ),
        ResultPath=path_converter("$.detail.shipped"),
        End=True,
    )

    _load_and_test_vector(kind="map", name=name, value=test)



"""Custom attrs validators."""


def _validate_service_integrations(service_name, api_name):
    # type: (str, str) -> bool
    # TODO: Actually do this.
    return True


def is_valid_arn(value):
    # type: (str) -> bool
    try:
        arn_literal, _partition, service, _region, _account, resource_type, name = value.split(":", 6)
    except ValueError:
        return False

    return _validate_service_integrations(service_name=resource_type, api_name=name)


def is_valid_timestamp(value):
    # type: (str) -> bool
    # TODO: Actually do this.
    return True


_CHOICE_COMPARISON_OPERATORS = (
    "StringEquals",
    "StringLessThan",
    "StringGreaterThan",
    "StringLessThanEquals",
    "StringGreaterThanEquals",
    "NumericEquals",
    "NumericLessThan",
    "NumericGreaterThan",
    "NumericLessThanEquals",
    "NumericGreaterThanEquals",
    "BooleanEquals",
    "TimestampEquals",
    "TimestampLessThan",
    "TimestampGreaterThan",
    "TimestampLessThanEquals",
    "TimestampGreaterThanEquals",
    "And",
    "Or",
    "Not",
)



"""rhodes."""
import io
import os
import re

from setuptools import find_packages, setup

VERSION_RE = re.compile(r"""__version__ = ['"]([0-9b.]+)['"]""")
HERE = os.path.abspath(os.path.dirname(__file__))


def read(*args):
    """Read complete file contents."""
    return io.open(os.path.join(HERE, *args), encoding="utf-8").read()


def get_version():
    """Read the version from this module."""
    init = read("src", "rhodes", "__init__.py")
    return VERSION_RE.search(init).group(1)


def get_requireemnts():
    """Read the requirements file."""
    raw_requirements = read("requirements.txt")
    requirements = []
    dependencies = []

    for req in raw_requirements.splitlines():
        req = req.strip()
        if not req:
            continue
        elif req.startswith("#"):
            continue
        elif "+" in req:
            dependencies.append(req)
        else:
            requirements.append(req)

    return requirements, dependencies


INSTALL_REQUIRES, DEPENDENCY_LINKS = get_requireemnts()

setup(
    name="rhodes",
    version=get_version(),
    packages=find_packages("src"),
    package_dir={"": "src"},
    url="https://github.com/mattsb42/rhodes",
    author="Matt Bullock",
    author_email="m@ttsb42.com",
    maintainer="Matt Bullock",
    description="rhodes",
    long_description=read("README.rst"),
    keywords="rhodes rhodes",
    data_files=["README.rst", "CHANGELOG.rst", "LICENSE", "requirements.txt"],
    license="Apache 2.0",
    install_requires=INSTALL_REQUIRES,
    dependency_links=DEPENDENCY_LINKS,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
)



"""Unit test suite for ``rhodes.states.services.awslambda``."""
import pytest

from rhodes.states import Parameters, Task
from rhodes.states.services import IntegrationPattern
from rhodes.states.services.awslambda import AwsLambda
from rhodes.structures import JsonPath

pytestmark = [pytest.mark.local, pytest.mark.functional]


def test_aws_lambda_request_response():
    expected_task = Task(
        "TestTask",
        Resource="arn:aws:states:::lambda:invoke",
        Parameters=Parameters(
            FunctionName="arn:aws:lambda:us-east-1:123456789012:function:ship-val",
            Payload={"test": "value"},
            InvocationType="RequestResponse",
        ),
    )
    expected = expected_task.to_dict()

    test_task = AwsLambda(
        "TestTask", FunctionName="arn:aws:lambda:us-east-1:123456789012:function:ship-val", Payload={"test": "value"}
    )
    test = test_task.to_dict()

    assert test == expected


def test_aws_lambda_callback():
    expected_task = Task(
        "TestTask",
        Resource="arn:aws:states:::lambda:invoke.waitForTaskToken",
        Parameters=Parameters(
            FunctionName="arn:aws:lambda:us-east-1:123456789012:function:ship-val",
            Payload={"test": "value"},
            InvocationType="RequestResponse",
        ),
    )
    expected = expected_task.to_dict()

    test_task = AwsLambda(
        "TestTask",
        Pattern=IntegrationPattern.WAIT_FOR_CALLBACK,
        FunctionName="arn:aws:lambda:us-east-1:123456789012:function:ship-val",
        Payload={"test": "value"},
    )
    test = test_task.to_dict()

    assert test == expected



"""Placeholder module to remind you to write tests."""
import pytest

pytestmark = [pytest.mark.local, pytest.mark.functional]


@pytest.mark.xfail(strict=True)
def test_write_tests():
    assert False



"""AWS Step Functions Integrations for Task states.

https://docs.aws.amazon.com/step-functions/latest/dg/connect-stepfunctions.html
"""
import attr

from rhodes._util import RequiredValue
from rhodes.states.services import IntegrationPattern, ServiceArn, ServiceIntegration, _supports_patterns


@attr.s(eq=False)
@_supports_patterns(
    IntegrationPattern.REQUEST_RESPONSE, IntegrationPattern.SYNCHRONOUS, IntegrationPattern.WAIT_FOR_CALLBACK
)
class AwsStepFunctions(ServiceIntegration):
    _required_fields = (RequiredValue("stateMachineArn", "AWS Step Functions Task requires a state machine target"),)
    _resource_name = ServiceArn.STEP_FUNCTIONS

    # TODO: Sort out validation rules
    #  https://docs.aws.amazon.com/step-functions/latest/apireference/API_StartExecution.html#API_StartExecution_RequestSyntax

    # SFn docs say that this needs to be a string,
    #  but in practice JSON can be provided inline in the state machine
    input = attr.ib(default=None)
    name = attr.ib(default=None)
    stateMachineArn = attr.ib(default=None)



"""Undifferentiated unit tests for ``rhodes.states``."""
import pytest

from rhodes.states import Parameters

pytestmark = [pytest.mark.local, pytest.mark.functional]


def test_parameters_repr():
    test = Parameters(a="A", b=3, c=True)

    assert repr(test) == "Parameters(a='A', b=3, c=True)"



"""Amazon SQS Integrations for Task states.

https://docs.aws.amazon.com/step-functions/latest/dg/connect-sqs.html
"""
import attr

from rhodes._util import RequiredValue
from rhodes.states.services import IntegrationPattern, ServiceArn, ServiceIntegration, _supports_patterns


@attr.s(eq=False)
@_supports_patterns(IntegrationPattern.REQUEST_RESPONSE, IntegrationPattern.WAIT_FOR_CALLBACK)
class AmazonSqs(ServiceIntegration):
    _required_fields = (
        RequiredValue("MessageBody", "Amazon SQS Task requires a message body"),
        RequiredValue("QueueUrl", "Amazon SQS Task requires a queue url"),
    )
    _resource_name = ServiceArn.SQS

    # TODO: Sort out validation rules
    #  https://docs.aws.amazon.com/step-functions/latest/dg/connect-sqs.html

    DelaySeconds = attr.ib(default=None)
    MessageAttribute = attr.ib(default=None)
    MessageBody = attr.ib(default=None)
    MessageDeduplicationId = attr.ib(default=None)
    MessageGroupId = attr.ib(default=None)
    QueueUrl = attr.ib(default=None)



"""Amazon SageMaker Integrations for Task states.

https://docs.aws.amazon.com/step-functions/latest/dg/connect-sagemaker.html
"""
import attr

from rhodes._util import RequiredValue
from rhodes.states.services import IntegrationPattern, ServiceArn, ServiceIntegration, _supports_patterns

__all__ = ("AmazonSageMaker",)

_sagemaker_supports_patterns = _supports_patterns(IntegrationPattern.REQUEST_RESPONSE, IntegrationPattern.SYNCHRONOUS)


def _tags(cls: ServiceIntegration) -> ServiceIntegration:
    cls.Tags = attr.ib(default=None)

    return cls


def _endpoint_config_name(cls: ServiceIntegration) -> ServiceIntegration:
    cls.EndpointConfigName = attr.ib(default=None)

    return cls


def _endpoint_name(cls: ServiceIntegration) -> ServiceIntegration:
    cls.EndpointName = attr.ib(default=None)

    return cls


@attr.s(eq=False)
@_sagemaker_supports_patterns
class AmazonSageMaker(ServiceIntegration):
    pass


@attr.s(eq=False)
@_sagemaker_supports_patterns
@_tags
@_endpoint_name
@_endpoint_config_name
class AmazonSageMakerCreateEndpoint(ServiceIntegration):
    _required_fields = ()
    _resource_name = ServiceArn.SAGEMAKER_CREATE_ENDPOINT

    # TODO: Sort out validation rules
    #  https://docs.aws.amazon.com/sagemaker/latest/dg/API_CreateEndpoint.html#API_CreateEndpoint_RequestParameters


@attr.s(eq=False)
@_sagemaker_supports_patterns
@_tags
@_endpoint_config_name
class AmazonSageMakerCreateEndpointConfig(ServiceIntegration):
    _required_fields = ()
    _resource_name = ServiceArn.SAGEMAKER_CREATE_ENDPOINT_CONFIG

    # TODO: Sort out validation rules
    #  https://docs.aws.amazon.com/sagemaker/latest/dg/API_CreateEndpointConfig.html#API_CreateEndpointConfig_RequestParameters

    KmsKeyId = attr.ib(default=None)
    ProductionVariants = attr.ib(default=None)


@attr.s(eq=False)
@_sagemaker_supports_patterns
@_tags
class AmazonSageMakerCreateHyperParameterTuningJob(ServiceIntegration):
    _required_fields = ()
    _resource_name = ServiceArn.SAGEMAKER_CREATE_HYPER_PARAMETER_TUNING_JOB

    # TODO: Sort out validation rules
    #  https://docs.aws.amazon.com/sagemaker/latest/dg/API_CreateHyperParameterTuningJob.html#API_CreateHyperParameterTuningJob_RequestParameters

    HyperParameterTuningJobConfig = attr.ib(default=None)
    HyperParameterTuningJobName = attr.ib(default=None)
    TrainingJobDefinition = attr.ib(default=None)
    WarmStartConfig = attr.ib(default=None)


@attr.s(eq=False)
@_sagemaker_supports_patterns
@_tags
class AmazonSageMakerCreateLabelingJob(ServiceIntegration):
    _required_fields = ()
    _resource_name = ServiceArn.SAGEMAKER_CREATE_LABELING_JOB

    # TODO: Sort out validation rules
    #  https://docs.aws.amazon.com/sagemaker/latest/dg/API_CreateLabelingJob.html#API_CreateLabelingJob_RequestParameters

    HumanTaskConfig = attr.ib(default=None)
    InputConfig = attr.ib(default=None)
    LabelAttributeName = attr.ib(default=None)
    LabelCategoryConfigS3Uri = attr.ib(default=None)
    LabelingJobAlgorithmsConfig = attr.ib(default=None)
    LabelingJobName = attr.ib(default=None)
    OutputConfig = attr.ib(default=None)
    RoleArn = attr.ib(default=None)
    StoppingConditions = attr.ib(default=None)


@attr.s(eq=False)
@_sagemaker_supports_patterns
@_tags
class AmazonSageMakerCreateModel(ServiceIntegration):
    _required_fields = ()
    _resource_name = ServiceArn.SAGEMAKER_CREATE_MODEL

    # TODO: Sort out validation rules
    #  https://docs.aws.amazon.com/sagemaker/latest/dg/API_CreateModel.html#API_CreateModel_RequestParameters

    Containers = attr.ib(default=None)
    EnableNetworkIsolation = attr.ib(default=None)
    ExecutionRoleArn = attr.ib(default=None)
    ModelName = attr.ib(default=None)
    PrimaryContainer = attr.ib(default=None)
    VpcConfig = attr.ib(default=None)


@attr.s(eq=False)
@_sagemaker_supports_patterns
@_tags
class AmazonSageMakerCreateTrainingJob(ServiceIntegration):
    _required_fields = ()
    _resource_name = ServiceArn.SAGEMAKER_CREATE_TRAINING_JOB

    # TODO: Sort out validation rules
    #  https://docs.aws.amazon.com/sagemaker/latest/dg/API_CreateTrainingJob.html#API_CreateTrainingJob_RequestParameters

    AlgorithmSpecification = attr.ib(default=None)
    HyperParameters = attr.ib(default=None)
    InputDataConfig = attr.ib(default=None)
    OutputDataConfig = attr.ib(default=None)
    ResourceConfig = attr.ib(default=None)
    RoleArn = attr.ib(default=None)
    StoppingCondition = attr.ib(default=None)
    TrainingJobName = attr.ib(default=None)
    VpcConfig = attr.ib(default=None)


@attr.s(eq=False)
@_sagemaker_supports_patterns
@_tags
class AmazonSageMakerCreateTransformJob(ServiceIntegration):
    _required_fields = ()
    _resource_name = ServiceArn.SAGEMAKER_CREATE_TRANSFORM_JOB

    # TODO: Sort out validation rules
    #  https://docs.aws.amazon.com/sagemaker/latest/dg/API_CreateTransformJob.html#API_CreateTransformJob_RequestParameters

    BatchStrategy = attr.ib(default=None)
    Environment = attr.ib(default=None)
    MaxConcurrentTransforms = attr.ib(default=None)
    MaxPayloadInMB = attr.ib(default=None)
    ModelName = attr.ib(default=None)
    TransformInput = attr.ib(default=None)
    TransformJobName = attr.ib(default=None)
    TransformOutput = attr.ib(default=None)
    TransformResources = attr.ib(default=None)


@attr.s(eq=False)
@_sagemaker_supports_patterns
@_tags
@_endpoint_name
@_endpoint_config_name
class AmazonSageMakerUpdateEndpoint(ServiceIntegration):
    _required_fields = ()
    _resource_name = ServiceArn.SAGEMAKER_UPDATE_ENDPOINT

    # TODO: Sort out validation rules
    #  https://docs.aws.amazon.com/sagemaker/latest/dg/API_UpdateEndpoint.html#API_UpdateEndpoint_RequestParameters
